import ulab.numpy as np
import time
import os
import ujson
from time import *
import nncase_runtime as nn
import ulab.numpy as np
import image
import gc
import time


root_path="/sdcard/mp_deployment_source/"        # root_path要以/结尾
config_path=root_path+"deploy_config.json"
memory_path=root_path+"memory.bin"
image_path=root_path+"test.jpg"
debug_mode=1

class ScopedTiming:
    def __init__(self, info="", enable_profile=True):
        self.info = info
        self.enable_profile = enable_profile

    def __enter__(self):
        if self.enable_profile:
            self.start_time = time.time_ns()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.enable_profile:
            elapsed_time = time.time_ns() - self.start_time
            print(f"{self.info} took {elapsed_time / 1000000:.2f} ms")


def cdist_ulab_matmul(x1, x2, p=2.0):
    """
    计算两组数据之间的欧几里得距离。

    参数:
        x1 (ndarray): 第一组数据。
        x2 (ndarray): 第二组数据。
        p (float, 默认为2.0): 用于计算的范数。目前只支持欧几里得范数 (p=2.0)。

    返回:
        ndarray: x1 和 x2 之间的距离矩阵。
    """

    x1 = x1.reshape((-1, x1.shape[-1]))
    x2 = x2.reshape((-1, x2.shape[-1]))

    if p == 2.0:
        x1_norm = np.sum(x1 ** 2, axis=1)
        x2_norm = np.sum(x2 ** 2, axis=1)

        # 手动增加一个维度
        x1_norm = x1_norm.reshape((len(x1_norm), 1))
        x2_norm_T = x2_norm.reshape((1, len(x2_norm)))  # 这相当于转置


        x1x2 = np.dot(x1, x2.T)  # 如果ulab支持.T，就使用它来进行转置


        dist = np.sqrt(np.maximum(x1_norm + x2_norm_T - 2 * x1x2, 0))

    return dist


def topk(arr, k, axis=-1, largest=True):
    """
    找到数组中的前k个最大或最小元素。

    参数:
        arr (ndarray): 输入数组。
        k (int): 需要查找的元素数量。
        axis (int, 默认为-1): 指定的轴。
        largest (bool, 默认为True): 如果为True则找最大的k个元素，否则找最小的k个元素。

    返回:
        tuple: 一个包含两个ndarrays的元组：前k个元素的值和它们的索引。
    """
    sorted_indices = np.argsort(arr, axis=axis)

    if largest:
        idx_topk = sorted_indices[:, -k:]
    else:
        idx_topk = sorted_indices[:, :k]

    # 使用ulab的索引功能从原始数组中获取值
    if axis == 0:
        val_topk = arr[idx_topk]
    elif axis == 1:
        val_topk = np.array([[arr[i, idx_topk[i, j]] for j in range(k)] for i in range(arr.shape[0])])
    else:
        raise ValueError("Unsupported axis value.")

    return val_topk, idx_topk


def softmax(x):
    """
    对输入的ndarray执行softmax函数。

    参数:
        x (ndarray): 输入数组。

    返回:
        ndarray: softmax处理后的数组。
    """

    x_shifted = x - np.max(x)
    exp_x = np.exp(x_shifted)

    if len(x.shape) == 2:
        sum_exp_x = np.sum(exp_x, axis=1)
        for i in range(x.shape[0]):
            exp_x[i, :] = exp_x[i, :] / sum_exp_x[i]
        return exp_x
    else:
        return exp_x / np.sum(exp_x)


def nearest_neighbors(embedding, n_neighbors, memory_bank):
    """使用欧几里得范数求最近邻。

    参数:
        embedding (Tensor): 输入特征特征。
        n_neighbors (int): 考虑的最近邻居数量
        memory_bank (Tensor):memory存储向量

    返回:
        Tensor: patch分数。
        Tensor: 最近邻索引。
    """

    distances = cdist_ulab_matmul(embedding, memory_bank)
    if n_neighbors == 1:
        patch_scores = np.min(distances, 1)
        locations = np.argmin(distances,1)
    else:
        patch_scores, locations = topk(distances, n_neighbors, axis=1, largest=False)

    return patch_scores, locations


def compute_anomaly_score(patch_scores, locations, embedding, memory_bank, num_neighbors=9):
    """计算Image Level异常分数。

    参数:
        patch_scores (Tensor): patch异常分数。
        locations: 对于每个patch位置，memory中最近邻居的位置。
        embedding: 生成patch分数的特征embedding。
    返回:
        Tensor: Image Level异常分数。
    """

    if num_neighbors == 1:
        return np.max(patch_scores, 1)

    batch_size, num_patches = patch_scores.shape

    max_patches = np.argmax(patch_scores, 1)

    # 形状变换并索引embedding数组
    reshaped_embedding = embedding.reshape((batch_size, num_patches, -1))


    max_patches_features = np.array([reshaped_embedding[0, max_patches, :]])

    score = patch_scores[0, max_patches]

    nn_index = locations[0, max_patches]

    nn_sample = memory_bank[nn_index, :]
    nn_sample = nn_sample.reshape((1,-1))

    _, support_samples = nearest_neighbors(nn_sample, memory_bank=memory_bank, n_neighbors=num_neighbors)

    # 使用循环从memory中索引样本
    gathered_samples_list = []
    for indices in support_samples[0]:
        gathered_samples_list.append(memory_bank[indices, :])
    gathered_samples = np.array(gathered_samples_list)

    distances = cdist_ulab_matmul(max_patches_features, gathered_samples, p=2.0)

    weights = 1 - softmax(distances[0])[0]


    score = weights * score

    del gathered_samples
    del nn_sample
    del max_patches_features
    del reshaped_embedding
    return score



def nearest_neighbor_interpolation(input_image, target_size):
    """
    使用最近邻插值对输入图片进行缩放。

    参数:
        input_image (numpy.ndarray): 输入图片，其形状为 (batch_size, channels, height, width)。
        target_size (tuple): 目标大小，一个二元组 (height, width)。

    返回:
        numpy.ndarray: 缩放后的图片。
    """
    import time

    batch_size, C, H, W = input_image.shape

    output_image = np.zeros((batch_size, C, target_size[0], target_size[1]))

    row_scale = H / target_size[0]
    col_scale = W / target_size[1]

    for b in range(batch_size):
        for c in range(C):
            for i in range(target_size[0]):
                src_i = int(i * row_scale)
                for j in range(target_size[1]):
                    src_j = int(j * col_scale)
                    output_image[b, c, i, j] = input_image[b, c, src_i, src_j]



    return output_image

def avg_pool_forward_ulab(x, kernel_size=2, stride=2, padding=0):
    """
    使用uLab执行平均池化操作。

    参数:
        x (numpy.ndarray): 输入数据，其形状为 (N, C, H, W)，其中N是批次大小，C是通道数，H和W分别是高和宽。
        kernel_size (int, 可选): 池化核的大小。默认为2。
        stride (int, 可选): 池化操作的步长。默认为2。
        padding (int, 可选): 在输入数据的高和宽周围添加的零填充的数量。默认为0。

    返回:
        numpy.ndarray: 经过平均池化操作的输出数据。
    """
    N, C, H, W = x.shape

    # 模拟padding
    if padding > 0:
        temp = np.zeros((N, C, H + 2 * padding, W + 2 * padding))
        temp[:, :, padding:padding+H, padding:padding+W] = x
        x = temp
        del temp

    out_h = (H + 2*padding - kernel_size) // stride + 1
    out_w = (W + 2*padding - kernel_size) // stride + 1

    out = np.zeros((N, C, out_h, out_w))

    for i in range(out_h):
        for j in range(out_w):
            h_start = i * stride
            h_end = h_start + kernel_size
            w_start = j * stride
            w_end = w_start + kernel_size

            # 使用索引进行平均池化操作
            pooled = np.mean(x[:, :, h_start:h_end, w_start:w_end], axis=2)
            out[:, :, i, j] = np.mean(pooled, axis=2)
            del pooled

    return out


def generate_embedding(features):
    """
    从输入特征图中生成embedding。

    参数:
        features (dict): 输入特征

    返回:
        ulab.array: 嵌入向量。
    """
    layers = ['layer2','layer3']
    embeddings = features[layers[0]]

    for layer in layers[1:]:
        layer_embedding = features[layer]

        # 使用最近邻插值来调整图层的大小以匹配第一个图层

        layer_embedding = nearest_neighbor_interpolation(layer_embedding, (embeddings.shape[2], embeddings.shape[3]))

        # 使用concatenate进行数组连接
        embeddings = np.concatenate((embeddings, layer_embedding), axis=1)
        del layer_embedding

    return embeddings


def reshape_embedding(embedding):
    """
    使用ulab对嵌入向量进行重塑。

    将嵌入向量从以下格式:
    [Batch, Embedding, Patch, Patch] 重塑为 [Batch*Patch*Patch, Embedding]。

    参数:
        embedding (ndarray): 形状为 [Batch, Embedding, Patch, Patch] 的嵌入张量。

    返回:
        ndarray: 形状为 [Batch*Patch*Patch, Embedding] 的重塑后的嵌入张量。
    """
    Batch, Embedding, Patch, _ = embedding.shape
    reshaped = embedding.reshape((Embedding, Patch * Patch))
    reshaped = reshaped.transpose()
    reshaped_return = reshaped.copy()
    reshaped_return = reshaped_return.reshape((Patch * Patch, Embedding))

    del reshaped
    return reshaped_return


def read_img(img_path):
    img_data = image.Image(img_path)
    img_data_rgb888=img_data.to_rgb888()
    img_hwc=img_data_rgb888.to_numpy_ref()
    shape=img_hwc.shape
    img_tmp = img_hwc.reshape((shape[0] * shape[1], shape[2]))
    img_tmp_trans = img_tmp.transpose()
    img_res=img_tmp_trans.copy()
    img_return=img_res.reshape((shape[2],shape[0],shape[1]))
    return img_return

# 读取deploy_config.json文件
def read_deploy_config(config_path):
    # 打开JSON文件以进行读取deploy_config
    with open(config_path, 'r') as json_file:
        try:
            # 从文件中加载JSON数据
            config = ujson.load(json_file)
            # 打印数据（可根据需要执行其他操作）
            #print(config)
        except ValueError as e:
            print("JSON 解析错误:", e)
    return config

# 任务后处理
def softmax(x):
    exp_x = np.exp(x - np.max(x))
    return exp_x / np.sum(exp_x)

def anomaly_detection():
    print("--------------start-----------------")
    # 使用json读取内容初始化部署变量
    x = np.fromfile(memory_path,dtype=np.float)
    deploy_conf=read_deploy_config(config_path)
    kmodel_name=deploy_conf["kmodel_path"]
    image_threshold=deploy_conf["image_threshold"]
    img_size=deploy_conf["img_size"]
    memory_shape = deploy_conf["memory_bank_shape"]
    ano_res=""
    memory_bank = x.reshape((memory_shape[0],memory_shape[1]))
    # ai2d输入输出初始化
    ai2d_input = read_img(image_path)
    ai2d_input_tensor = nn.from_numpy(ai2d_input)
    ai2d_input_shape=ai2d_input.shape
    data = np.ones((1,3,img_size[0],img_size[1]),dtype=np.uint8)
    ai2d_out = nn.from_numpy(data)
    kpu = nn.kpu()
    ai2d = nn.ai2d()
    kpu.load_kmodel(root_path+kmodel_name)
    ai2d.set_dtype(nn.ai2d_format.NCHW_FMT,
                                   nn.ai2d_format.NCHW_FMT,
                                   np.uint8, np.uint8)
    ai2d.set_resize_param(True, nn.interp_method.tf_bilinear, nn.interp_mode.half_pixel )
    ai2d_builder = ai2d.build([1,3,ai2d_input_shape[1],ai2d_input_shape[2]], [1,3,img_size[0],img_size[1]])
    with ScopedTiming("total",debug_mode > 0):
        ai2d_builder.run(ai2d_input_tensor, ai2d_out)
        kpu.set_input_tensor(0, ai2d_out)
        kpu.run()
        del ai2d_input_tensor
        del ai2d_out
        data_0 = kpu.get_output_tensor(0)
        eb_0 = data_0.to_numpy()
        data_1 = kpu.get_output_tensor(1)
        eb_1 = data_1.to_numpy()
        features = {'layer2': avg_pool_forward_ulab(eb_0, 3, 1, 1),
                    'layer3': avg_pool_forward_ulab(eb_1, 3, 1, 1)
                    }
        embedding = generate_embedding(features)
        embedding_result = reshape_embedding(embedding)

        patch_scores, locations = nearest_neighbors(embedding=embedding_result, n_neighbors=1, memory_bank=memory_bank)
        patch_scores = patch_scores.reshape((1, -1))  # [1, 1024]
        locations = locations.reshape((1, -1))
        anomaly_score = compute_anomaly_score(patch_scores, locations, embedding=embedding_result, memory_bank=memory_bank)
        if anomaly_score > image_threshold:
            print("input image is anomaly")
            ano_res="anomaly"
        else:
            print("input image is nomal")
            ano_res="nomal"
        image_draw=image.Image(image_path).to_rgb565()
        image_draw.draw_string(10, 10,ano_res , scale=2,color=(255,0,255,0))
        image_draw.compress_for_ide()
        image_draw.save(root_path+"ano_result.jpg")
        del data_0
        del data_1

    del ai2d
    del ai2d_builder
    del kpu
    print("---------------end------------------")
    gc.collect()
    nn.shrink_memory_pool()

if __name__ == '__main__':
    nn.shrink_memory_pool()
    anomaly_detection()



